from functools import reduce
import re
from typing import Any, Iterable, Optional, Tuple, Type, Union, cast

import decimal
from decimal import Decimal, ROUND_HALF_UP

from .currency import BaseCurrency
from .exceptions import CurrencyMismatchError, ConversionError, InvalidOperandError


__all__ = ["Money"]

DEFAULT_MIN_DECIMALS = 2
DEFAULT_MAX_DECIMALS = 9
UNITS_MAX_LENGTH = 18
NANOS_LENGTH = 9

HIGHEST_SUPPORTED_AMOUNT = "999999999999999999.999999999"
LOWEST_SUPPORTED_AMOUNT = "-999999999999999999.999999999"

RoundingContext = decimal.Context(rounding=ROUND_HALF_UP)

_parse_format_specifier_regex = re.compile(
    r"""\A
(?:
   (?P<fill>.)?
   (?P<align>[<>=^])
)?
(?P<sign>[-+ ])?
(?P<zeropad>0)?
(?P<minimumwidth>(?!0)\d+)?
(?P<thousands_sep>[,_])?
(?:\.(?P<precision>0|(?!0)\d+))?
(?P<type>[cCdfFmMs])?
\Z
""",
    re.VERBOSE,
)


class DefaultCurrency:
    pass


class Money:
    __slots__ = ("_amount", "_currency")
    _amount: Decimal
    _currency: Optional[Union[BaseCurrency, str]]

    @classmethod
    def sort(cls, iterable: Iterable, reverse: bool = False) -> Iterable:
        return sorted(iterable, key=lambda x: x if isinstance(x, Money) else Money(x), reverse=reverse)

    @classmethod
    def sum(
        cls,
        iterable: Iterable,
        currency: Optional[Union[Type[DefaultCurrency], BaseCurrency, str]] = DefaultCurrency,
        from_sub_units: Optional[bool] = None,
        **kwargs: Any,
    ) -> "Money":
        return reduce(
            lambda v, e: v + (e if isinstance(e, Money) else Money(e, from_sub_units=from_sub_units)),
            iterable,
            Money(0, currency=currency, from_sub_units=from_sub_units),
        )

    @classmethod
    def _is_unknown_amount_type(cls, amount: Optional[Union["Money", Decimal, int, float, str, object]]) -> bool:
        return not any(map(lambda type_: isinstance(amount, type_), (Money, Decimal, int, bool, float, str)))

    @classmethod
    def from_sub_units(
        cls,
        amount: Optional[Union["Money", Decimal, int, float, str, object]],
        currency: Optional[Union[Type[DefaultCurrency], BaseCurrency, str]] = DefaultCurrency,
    ) -> "Money":
        return Money(amount=amount, currency=currency, from_sub_units=True)

    def __init__(
        self,
        amount: Optional[Union["Money", Decimal, int, float, str, object]] = None,
        currency: Optional[Union[Type[DefaultCurrency], BaseCurrency, str]] = DefaultCurrency,
        from_sub_units: Optional[bool] = None,
        units: Optional[int] = None,
        nanos: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        validate_amounts = []

        if units is not None or nanos is not None:
            try:
                units = units or 0
                nanos = nanos or 0
                if (units > 0 and nanos < 0) or (units < 0 and nanos > 0):
                    raise ValueError
                units_str = str(units).lstrip("-")
                nanos_str = str(nanos).lstrip("-").rjust(NANOS_LENGTH, "0")
                if len(units_str) > UNITS_MAX_LENGTH:
                    raise ValueError
                if len(nanos_str) != NANOS_LENGTH:
                    raise ValueError
                sign = "-" if nanos < 0 or units < 0 else ""
                new_amount = Decimal(f"{sign}{units_str}.{nanos_str}")
                if amount is None:
                    amount = new_amount
                else:
                    validate_amounts.append(new_amount)
            except Exception:
                raise ConversionError("Invalid values for units and nanos")

        if amount is None:
            raise ConversionError("Missing input values for monetary amount")

        if (
            isinstance(amount, Money)
            and currency is DefaultCurrency
            and from_sub_units is None
            and units is None
            and nanos is None
        ):
            object.__setattr__(self, "_amount", amount._amount)
            object.__setattr__(self, "_currency", amount._currency)
            return

        if (
            currency is not DefaultCurrency
            and not isinstance(currency, str)
            and not isinstance(currency, BaseCurrency)
            and currency is not None
        ):
            raise ConversionError("Invalid currency value")

        output_amount = None
        output_currency: Optional[Union[BaseCurrency, str]] = None
        if currency is not DefaultCurrency:
            if isinstance(currency, BaseCurrency):
                output_currency = currency
            else:
                output_currency = str(currency or "").strip() or None
                output_currency = (
                    output_currency.upper() if output_currency and len(output_currency) == 3 else output_currency
                )

        if Money._is_unknown_amount_type(amount):
            try:
                match_amount = getattr(amount, "amount")
                match_amount = (match_amount()) if match_amount and callable(match_amount) else match_amount
                if match_amount is None or Money._is_unknown_amount_type(match_amount):
                    raise AttributeError

                match_currency = None
                try:
                    match_currency = getattr(amount, "currency")
                    match_currency = (
                        (match_currency()) if match_currency and callable(match_currency) else match_currency
                    )
                    if not match_currency:
                        raise AttributeError
                except AttributeError:
                    matches = re.match(r"^(?:[-+]?[0-9.]+)[ ]+([a-zA-Z]+)$", str(amount))
                    if not matches:
                        matches = re.match(r"^([a-zA-Z]+)[ ]+(?:[-+]?[0-9.]+)$", str(amount))
                    if matches:
                        match_currency = matches.group(1)

                if match_currency is not None:
                    match_currency = str(match_currency).strip()
                    match_currency = (
                        match_currency.upper()
                        if match_currency and isinstance(match_currency, str) and len(match_currency) == 3
                        else match_currency
                    )
                    if output_currency is not None and match_currency != output_currency:
                        raise ConversionError("Mismatching currency in input value and currency argument")
                    output_currency = output_currency if isinstance(output_currency, BaseCurrency) else match_currency

                amount = match_amount
            except AttributeError:
                amount = str(amount)

        if amount is not None and isinstance(amount, int) and not isinstance(amount, bool):
            output_amount = Decimal(amount)
        elif amount is not None and isinstance(amount, float):
            output_amount = Decimal(str(amount))
        elif amount is not None and isinstance(amount, str) and amount.strip():
            amount = amount.strip()
            match_currency = None

            matches = re.match(r"^(?P<amount>[-+]?[0-9.]+)[ ]+(?P<currency>[a-zA-Z]+)$", amount)
            if not matches:
                matches = re.match(r"^(?P<currency>[a-zA-Z]+)[ ]+(?P<amount>[-+]?[0-9.]+)$", amount)
            if matches:
                amount = matches.group("amount").strip()
                match_currency = matches.group("currency").strip()
                match_currency = (
                    match_currency.upper()
                    if match_currency and isinstance(match_currency, str) and len(match_currency) == 3
                    else match_currency
                )

            if match_currency is not None:
                if output_currency is not None and match_currency != output_currency:
                    raise ConversionError("Mismatching currency in input value and currency argument")
                output_currency = output_currency if isinstance(output_currency, BaseCurrency) else match_currency

            try:
                output_amount = Decimal(amount)
            except Exception:
                raise ConversionError("Value cannot be used as monetary amount")
        elif amount is not None and isinstance(amount, Money):
            if amount.currency and not output_currency and currency is DefaultCurrency:
                output_currency = amount.currency

            output_amount = amount._amount
        elif amount is not None and isinstance(amount, Decimal):
            output_amount = amount

        if output_amount is None:
            raise ConversionError("Missing input values for monetary amount")

        if output_amount.is_infinite():
            raise ConversionError("Monetary amounts cannot be infinite")

        if output_amount.is_nan():
            raise ConversionError("Input amount is not a number")

        if from_sub_units:
            if output_currency and isinstance(output_currency, BaseCurrency):
                if output_currency.decimal_digits != 0:
                    output_amount = output_amount / Decimal(pow(10, output_currency.decimal_digits))
            else:
                output_amount = output_amount / 100

        if output_amount > Decimal(HIGHEST_SUPPORTED_AMOUNT):
            raise ConversionError(f"Input amount is too high, max value is {HIGHEST_SUPPORTED_AMOUNT}")

        if output_amount < Decimal(LOWEST_SUPPORTED_AMOUNT):
            raise ConversionError(f"Input amount is too low, min value is {LOWEST_SUPPORTED_AMOUNT}")

        if output_currency and not re.match(r"^[A-Za-z]+$", str(output_currency)):
            raise ConversionError("Invalid currency")

        if output_amount == 0 and output_amount.is_signed():
            output_amount = Decimal(0)

        if any([output_amount != a for a in validate_amounts]):
            raise ConversionError("Input arguments does not match")

        object.__setattr__(self, "_amount", output_amount)
        object.__setattr__(self, "_currency", output_currency)

    @property
    def amount(self) -> Decimal:
        return self._amount

    @property
    def currency(self) -> Optional[Union[BaseCurrency, str]]:
        return self._currency

    @property
    def _amount_tuple(self) -> Tuple[str, str]:
        amount = self._amount.quantize(Decimal(f"1e-{NANOS_LENGTH}"), ROUND_HALF_UP)
        sign, digits, exponent = amount.as_tuple()

        units_str = "".join(map(str, digits))[:exponent] or "0"
        nanos_str = "".join(map(str, digits))[exponent:]
        nanos_str = nanos_str.rjust((-exponent), "0").ljust(NANOS_LENGTH, "0")[0:NANOS_LENGTH]

        if sign and int(units_str):
            units_str = f"-{units_str}"
        if sign and int(nanos_str):
            nanos_str = f"-{nanos_str}"

        return units_str, nanos_str

    @property
    def units(self) -> int:
        units, _ = self._amount_tuple
        return int(units)

    @property
    def nanos(self) -> int:
        _, nanos = self._amount_tuple
        return int(nanos)

    def as_string(self, *args: Any, **kwargs: Any) -> str:
        amount = self.amount_as_string(*args, **kwargs)

        if self._currency:
            return f"{amount} {self._currency}"
        return str(amount)

    def as_str(self, *args: Any, **kwargs: Any) -> str:
        return self.as_string(*args, **kwargs)

    def as_decimal(self) -> Decimal:
        return self._amount

    def as_int(self) -> int:
        return int(self)

    def as_float(self) -> float:
        return float(self)

    def is_signed(self) -> bool:
        return self._amount.is_signed()

    def is_zero(self) -> bool:
        return self._amount == 0

    def add(self, other: Any, from_sub_units: Optional[bool] = None) -> "Money":
        return self + Money(other, from_sub_units=from_sub_units)

    def subtract(self, other: Any, from_sub_units: Optional[bool] = None) -> "Money":
        return self - Money(other, from_sub_units=from_sub_units)

    def sub(self, other: Any, from_sub_units: Optional[bool] = None) -> "Money":
        return self.subtract(other, from_sub_units=from_sub_units)

    def to_currency(self, currency: Optional[Union[BaseCurrency, str]]) -> "Money":
        return Money(self, currency=currency)

    def to(self, currency: Optional[Union[BaseCurrency, str]]) -> "Money":
        return self.to_currency(currency)

    def to_sub_units(self) -> "Money":
        if self._currency and isinstance(self._currency, BaseCurrency):
            if self._currency.decimal_digits == 0:
                return self
            return self * Decimal(pow(10, self._currency.decimal_digits))
        return self * 100

    def __setattr__(self, *args: Any) -> None:
        raise AttributeError("Attributes of monetary amounts cannot be changed")

    def __delattr__(self, *args: Any) -> None:
        raise AttributeError("Attributes of monetary amounts cannot be deleted")

    def amount_as_string(self, min_decimals: Optional[int] = None, max_decimals: Optional[int] = None) -> str:
        if min_decimals is None and max_decimals is None:
            if self._currency and isinstance(self._currency, BaseCurrency):
                min_decimals = self._currency.decimal_digits
            min_decimals = DEFAULT_MIN_DECIMALS if min_decimals is None else min_decimals
            max_decimals = max(min_decimals, DEFAULT_MAX_DECIMALS)
        elif min_decimals is None:
            if self._currency and isinstance(self._currency, BaseCurrency):
                min_decimals = self._currency.decimal_digits
            min_decimals = DEFAULT_MIN_DECIMALS if min_decimals is None else min_decimals
            min_decimals = min(min_decimals, max_decimals)
        elif max_decimals is None:
            max_decimals = max(min_decimals, DEFAULT_MAX_DECIMALS)

        min_decimals = cast(int, min_decimals)
        max_decimals = cast(int, max_decimals)
        if min_decimals > max_decimals:
            raise ValueError("Invalid values for min_decimals and max_decimals")

        amount = self._amount.quantize(Decimal(f"1e-{min_decimals}"), ROUND_HALF_UP)
        if amount == 0:
            amount = Decimal(0).quantize(Decimal(f"1e-{min_decimals}"))
        if max_decimals > min_decimals and amount != self._amount.quantize(
            Decimal(f"1e-{max_decimals}"), ROUND_HALF_UP
        ):
            amount = self._amount.quantize(Decimal(f"1e-{max_decimals}"), ROUND_HALF_UP)
            value = f"{amount:f}".rstrip("0")
        else:
            value = f"{amount:f}"
        return value

    def __repr__(self) -> str:
        return f'<stockholm.Money: "{self}">'

    def __str__(self) -> str:
        return self.as_string()

    def __format__(self, format_spec: str) -> str:
        if not format_spec:
            return str(self)

        m = _parse_format_specifier_regex.match(format_spec)
        if m is None:
            raise ValueError(f"Invalid format specifier: {format_spec}")

        format_dict = m.groupdict()

        fill = format_dict["fill"]
        align = format_dict["align"]
        zeropad = format_dict["zeropad"] is not None
        if zeropad:
            if fill is not None:
                raise ValueError(f"Fill character conflicts with '0' in format specifier: {format_spec}")
            if align is not None:
                raise ValueError(f"Alignment conflicts with '0' in format specifier: {format_spec}")
            fill = "0"
            align = ">"
        fill = fill or " "
        align = align or ">"

        sign = format_dict["sign"] or "-"

        minimumwidth = int(format_dict["minimumwidth"] or 0)
        precision = int(format_dict["precision"]) if format_dict["precision"] is not None else None

        thousands_sep = format_dict["thousands_sep"] if format_dict["thousands_sep"] is not None else ""

        output = ""

        if format_dict["type"] == "c":
            output = str(self._currency or "")

        if format_dict["type"] == "d":
            format_dict["type"] = "f"
            precision = 0

        with decimal.localcontext(RoundingContext):
            if format_dict["type"] in ("m", "M", "f", "F"):
                if precision is not None:
                    output = self.amount_as_string(min_decimals=precision, max_decimals=precision)
                else:
                    output = self.amount_as_string()

                output_amount = Money(output).amount

                if thousands_sep and (output_amount >= 1000 or output_amount <= -1000):
                    integral = int(output.split(".")[0])
                    try:
                        decimals = output.split(".")[1]
                        output = f"{integral:{thousands_sep}.0f}.{decimals}"
                    except IndexError:
                        output = f"{integral:{thousands_sep}.0f}"

                if format_dict["align"] == "=":
                    format_dict["align"] = ""
                    zeropad = True

                if zeropad:
                    if output.startswith("-"):
                        output = output[1:].rjust(minimumwidth - 1, fill)
                        output = f"-{output}"
                    else:
                        output = output.rjust(minimumwidth, fill)
                    if sign != "-" and output_amount >= 0:
                        if output.startswith(fill) and not output.startswith("0."):
                            output = f"{sign}{output[1:]}"
                        else:
                            output = f"{sign}{output}"
                elif sign != "-" and output_amount >= 0:
                    output = f"{sign}{output}"

                if not zeropad and not format_dict["align"] and align and fill and minimumwidth:
                    output = f"{output:{fill}{align}{minimumwidth}}"

                zeropad = False
                sign = ""
                thousands_sep = ""

                if self._currency:
                    if format_dict["type"] == "m":
                        output = f"{output} {self._currency}"
                    elif format_dict["type"] == "M":
                        output = f"{self._currency} {output}"
            elif format_dict["type"] in ("s", "", None):
                format_dict["type"] = "s"
                output = self.as_string()

        if sign and format_dict["sign"]:
            raise ValueError("Sign not allowed in string format specifier")
        if zeropad or format_dict["align"] == "=":
            raise ValueError("'=' alignment not allowed in string format specifier")
        if thousands_sep:
            raise ValueError("Cannot specify ',' in string format specifier")

        if format_dict["align"] and align and fill and minimumwidth:
            output = f"{output:{fill}{align}{minimumwidth}}"

        return output

    def __hash__(self) -> int:
        return hash(("stockholm.Money", self._amount, self._currency))

    def __bool__(self) -> bool:
        return bool(self._amount)

    def _convert_other(self, other: Any, allow_currency_mismatch: bool = False) -> "Money":
        if not isinstance(other, Money):
            try:
                converted_other = Money(other)
            except ConversionError as ex:
                other_repr = repr(other)
                self_repr = repr(self)
                raise InvalidOperandError(f"Unable to perform operations on {self_repr} with {other_repr}") from ex
        else:
            converted_other = other

        if (
            not allow_currency_mismatch
            and self._currency
            and converted_other._currency
            and self._currency != converted_other._currency
        ):
            raise CurrencyMismatchError("Unable to perform operations on values with differing currencies")

        return converted_other

    def _preferred_currency(self, other: "Money") -> Optional[Union[BaseCurrency, str]]:
        currency = self._currency if self._currency and isinstance(self._currency, BaseCurrency) else None
        currency = other._currency if not currency and other._currency and isinstance(other, BaseCurrency) else None
        return currency or self._currency or other._currency

    def __eq__(self, other: Any) -> bool:
        try:
            converted_other = self._convert_other(other, allow_currency_mismatch=True)
        except (ConversionError, InvalidOperandError):
            return False

        if self._currency and converted_other._currency and self._currency != converted_other._currency:
            if self._amount == 0 and converted_other._amount == 0:
                return True
            return False

        return self._amount == converted_other._amount

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __lt__(self, other: Any) -> bool:
        converted_other = self._convert_other(other)
        return self._amount < converted_other._amount

    def __le__(self, other: Any) -> bool:
        converted_other = self._convert_other(other)
        return self._amount <= converted_other._amount

    def __gt__(self, other: Any) -> bool:
        converted_other = self._convert_other(other)
        return self._amount > converted_other._amount

    def __ge__(self, other: Any) -> bool:
        converted_other = self._convert_other(other)
        return self._amount >= converted_other._amount

    def __add__(self, other: Any) -> "Money":
        converted_other = self._convert_other(other)
        amount = self._amount + converted_other._amount
        currency = self._preferred_currency(converted_other)
        return Money(amount, currency=currency)

    def __radd__(self, other: Any) -> "Money":
        return self.__add__(other)

    def __sub__(self, other: Any) -> "Money":
        converted_other = self._convert_other(other)
        amount = self._amount - converted_other._amount
        currency = self._preferred_currency(converted_other)
        return Money(amount, currency=currency)

    def __rsub__(self, other: Any) -> "Money":
        converted_other = self._convert_other(other)
        amount = converted_other._amount - self._amount
        currency = self._preferred_currency(converted_other)
        return Money(amount, currency=currency)

    def __mul__(self, other: Any) -> "Money":
        if not isinstance(other, Money):
            converted_other = self._convert_other(other)
        else:
            converted_other = other

        if converted_other._currency is not None and self._currency is not None:
            raise InvalidOperandError("Unable to multiply two monetary amounts with each other")

        amount = self._amount * converted_other._amount
        currency = self._preferred_currency(converted_other)
        return Money(amount, currency=currency)

    def __rmul__(self, other: Any) -> "Money":
        return self.__mul__(other)

    def __truediv__(self, other: Any) -> "Money":
        converted_other = self._convert_other(other, allow_currency_mismatch=True)

        if converted_other.amount == 0:
            raise ZeroDivisionError("division by zero")

        amount = self._amount / converted_other._amount

        if converted_other._currency is not None:
            return Money(amount, currency=None)

        return Money(amount, currency=self._currency)

    def __floordiv__(self, other: Any) -> "Money":
        converted_other = self._convert_other(other, allow_currency_mismatch=True)

        if converted_other.amount == 0:
            raise ZeroDivisionError("division by zero")

        amount = self._amount // converted_other._amount

        if converted_other._currency is not None:
            return Money(amount, currency=None)

        return Money(amount, currency=self._currency)

    def __mod__(self, other: Any) -> "Money":
        converted_other = self._convert_other(other, allow_currency_mismatch=True)
        amount = self._amount % converted_other._amount

        if self._currency == converted_other._currency:
            currency = self._preferred_currency(converted_other)
        else:
            currency = self._currency

        return Money(amount, currency=currency)

    def __divmod__(self, other: Any) -> Tuple["Money", "Money"]:
        converted_other = self._convert_other(other, allow_currency_mismatch=True)
        quotient, remainder = divmod(self._amount, converted_other._amount)

        if self._currency == converted_other._currency:
            currency = self._preferred_currency(converted_other)
        else:
            currency = self._currency

        if converted_other._currency is not None:
            return Money(quotient), Money(remainder, currency=currency)

        return Money(quotient, currency=currency), Money(remainder, currency=currency)

    def __pow__(self, other: Any) -> "Money":
        if not isinstance(other, Money):
            converted_other = self._convert_other(other)
        else:
            converted_other = other

        if converted_other._currency is not None:
            raise InvalidOperandError("Unable to use a monetary amount as an exponent")

        amount = self._amount ** converted_other._amount
        return Money(amount, currency=self._currency)

    def __neg__(self) -> "Money":
        return Money(-self._amount, currency=self._currency)

    def __pos__(self) -> "Money":
        return Money(+self._amount, currency=self._currency)

    def __abs__(self) -> "Money":
        return Money(abs(self._amount), currency=self._currency)

    def __int__(self) -> int:
        return int(self._amount)

    def __float__(self) -> float:
        return float(self._amount)

    def __round__(self, ndigits: int = 0) -> "Money":
        with decimal.localcontext(RoundingContext):
            amount = round(self._amount, ndigits)

        return Money(amount, currency=self._currency)



import sys
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union, cast


class MetaCurrency(type):
    ticker: str
    decimal_digits: int
    interchangeable_with: Optional[Union[Tuple[str, ...], List[str], Set[str]]]
    preferred_ticker: Optional[str]
    _meta: bool

    def __new__(cls, name: str, bases: Tuple[type, ...], attributedict: Dict) -> "MetaCurrency":
        ticker = attributedict.get("ticker", attributedict.get("__qualname__"))
        decimal_digits = attributedict.get("decimal_digits", 2)
        interchangeable_with = attributedict.get("interchangeable_with")
        preferred_ticker = attributedict.get("preferred_ticker")

        attributedict["ticker"] = ticker.split(".")[-1] if ticker else ""
        attributedict["currency"] = attributedict["ticker"]
        attributedict["decimal_digits"] = decimal_digits
        attributedict["interchangeable_with"] = sorted(interchangeable_with) if interchangeable_with else None
        attributedict["preferred_ticker"] = preferred_ticker if preferred_ticker else None

        attributedict["_meta"] = bool(not bases)
        attributedict["as_string"] = lambda: str(attributedict["ticker"])
        attributedict["as_str"] = lambda: str(attributedict["ticker"])

        result: Type[BaseCurrency] = type.__new__(cls, name, bases, attributedict)
        return result

    def __setattr__(self, *args: Any) -> None:
        raise AttributeError("Attributes of currencies cannot be changed")

    def __delattr__(self, *args: Any) -> None:
        raise AttributeError("Attributes of currencies cannot be deleted")

    def __repr__(self) -> str:
        if self._meta:
            return "<class 'stockholm.currency.Currency'>"
        return f'<stockholm.Currency: "{self}">'

    def __str__(self) -> str:
        if self._meta:
            return "<class 'stockholm.currency.Currency'>"
        return self.ticker or ""

    def __format__(self, format_spec: str) -> str:
        output = str(self)
        if format_spec.endswith("c"):
            format_spec = f"{format_spec[:-1]}s"
        return f"{output:{format_spec}}"

    def __eq__(self, other: Any) -> bool:
        if self.ticker:
            if not other:
                return False
            elif isinstance(other, BaseCurrency):
                return bool(self.ticker == other.ticker)
            elif isinstance(other, str):
                return bool(self.ticker == other)
        else:
            if isinstance(other, BaseCurrency):
                return not other.ticker
            elif isinstance(other, str):
                return bool(other == "")
        return False

    def __ne__(self, other: Any) -> bool:
        return not self == other

    @property  # type: ignore
    def __class__(self) -> Any:
        return BaseCurrency

    def __hash__(self) -> int:
        return hash(
            (
                "stockholm.MetaCurrency",
                self.ticker,
                self.decimal_digits,
                self.interchangeable_with,
                self.preferred_ticker,
            )
        )

    def __bool__(self) -> bool:
        return bool(self.ticker)


class BaseCurrency(metaclass=MetaCurrency):
    ticker: str
    decimal_digits: int
    interchangeable_with: Optional[Union[Tuple[str, ...], List[str], Set[str]]]
    preferred_ticker: Optional[str]
    _meta: bool

    def __init__(
        self,
        currency: Optional[Union["Currency", str]] = None,
        decimal_digits: Optional[int] = None,
        interchangeable_with: Optional[Union[Tuple[str, ...], List[str], Set[str]]] = None,
        preferred_ticker: Optional[str] = None,
    ) -> None:
        if currency and isinstance(currency, BaseCurrency):
            object.__setattr__(self, "ticker", currency.ticker)
            decimal_digits = currency.decimal_digits if decimal_digits is None else decimal_digits
            interchangeable_with = (
                currency.interchangeable_with if interchangeable_with is None else interchangeable_with
            )
            preferred_ticker = currency.preferred_ticker if preferred_ticker is None else preferred_ticker
        elif currency and isinstance(currency, str):
            object.__setattr__(self, "ticker", currency)
        else:
            object.__setattr__(self, "ticker", "")

        object.__setattr__(self, "currency", self.ticker)
        object.__setattr__(self, "decimal_digits", 2 if decimal_digits is None else decimal_digits)
        object.__setattr__(self, "interchangeable_with", sorted(interchangeable_with) if interchangeable_with else None)
        object.__setattr__(self, "preferred_ticker", preferred_ticker if preferred_ticker else None)

        object.__setattr__(self, "_meta", False)
        object.__setattr__(self, "as_string", lambda: str(self))
        object.__setattr__(self, "as_str", lambda: str(self))

    def __setattr__(self, *args: Any) -> None:
        raise AttributeError("Attributes of currencies cannot be changed")

    def __delattr__(self, *args: Any) -> None:
        raise AttributeError("Attributes of currencies cannot be deleted")

    def __repr__(self) -> str:
        return f'<stockholm.Currency: "{self.ticker}">'

    def __str__(self) -> str:
        return self.ticker or ""

    def __format__(self, format_spec: str) -> str:
        output = str(self)
        if format_spec.endswith("c"):
            format_spec = f"{format_spec[:-1]}s"
        return f"{output:{format_spec}}"

    def __eq__(self, other: Any) -> bool:
        if self.ticker:
            if not other:
                return False
            elif isinstance(other, BaseCurrency):
                return bool(self.ticker == other.ticker)
            elif isinstance(other, str):
                return bool(self.ticker == other)
        else:
            if isinstance(other, BaseCurrency):
                return not other.ticker
            elif isinstance(other, str):
                return bool(other == "")
        return False

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __hash__(self) -> int:
        return hash(
            ("stockholm.Currency", self.ticker, self.decimal_digits, self.interchangeable_with, self.preferred_ticker)
        )

    def __bool__(self) -> bool:
        return bool(self.ticker)


# ISO 4217 currency codes
class AED(BaseCurrency):
    pass


class AFN(BaseCurrency):
    pass


class ALL(BaseCurrency):
    pass


class AMD(BaseCurrency):
    pass


class ANG(BaseCurrency):
    pass


class AOA(BaseCurrency):
    pass


class ARS(BaseCurrency):
    pass


class AUD(BaseCurrency):
    pass


class AWG(BaseCurrency):
    pass


class AZN(BaseCurrency):
    pass


class BAM(BaseCurrency):
    pass


class BBD(BaseCurrency):
    pass


class BDT(BaseCurrency):
    pass


class BGN(BaseCurrency):
    pass


class BHD(BaseCurrency):
    pass


class BIF(BaseCurrency):
    decimal_digits = 2


class BMD(BaseCurrency):
    pass


class BND(BaseCurrency):
    pass


class BOB(BaseCurrency):
    pass


class BOV(BaseCurrency):
    pass


class BRL(BaseCurrency):
    pass


class BSD(BaseCurrency):
    pass


class BTN(BaseCurrency):
    pass


class BWP(BaseCurrency):
    pass


class BYN(BaseCurrency):
    pass


class BZD(BaseCurrency):
    pass


class CAD(BaseCurrency):
    pass


class CDF(BaseCurrency):
    pass


class CHE(BaseCurrency):
    pass


class CHF(BaseCurrency):
    pass


class CHW(BaseCurrency):
    pass


class CLF(BaseCurrency):
    decimal_digits = 4


class CLP(BaseCurrency):
    decimal_digits = 0


class CNY(BaseCurrency):
    interchangeable_with = ("CNH", "RMB")


class COP(BaseCurrency):
    pass


class COU(BaseCurrency):
    pass


class CRC(BaseCurrency):
    pass


class CUC(BaseCurrency):
    pass


class CUP(BaseCurrency):
    pass


class CVE(BaseCurrency):
    pass


class CZK(BaseCurrency):
    pass


class DJF(BaseCurrency):
    decimal_digits = 0


class DKK(BaseCurrency):
    pass


class DOP(BaseCurrency):
    pass


class DZD(BaseCurrency):
    pass


class EGP(BaseCurrency):
    pass


class ERN(BaseCurrency):
    pass


class ETB(BaseCurrency):
    pass


class EUR(BaseCurrency):
    pass


class FJD(BaseCurrency):
    pass


class FKP(BaseCurrency):
    pass


class GBP(BaseCurrency):
    pass


class GEL(BaseCurrency):
    pass


class GHS(BaseCurrency):
    pass


class GIP(BaseCurrency):
    pass


class GMD(BaseCurrency):
    pass


class GNF(BaseCurrency):
    decimal_digits = 0


class GTQ(BaseCurrency):
    pass


class GYD(BaseCurrency):
    pass


class HKD(BaseCurrency):
    pass


class HNL(BaseCurrency):
    pass


class HRK(BaseCurrency):
    pass


class HTG(BaseCurrency):
    pass


class HUF(BaseCurrency):
    pass


class IDR(BaseCurrency):
    pass


class ILS(BaseCurrency):
    interchangeable_with = ("NIS",)


class INR(BaseCurrency):
    pass


class IQD(BaseCurrency):
    decimal_digits = 3


class IRR(BaseCurrency):
    pass


class ISK(BaseCurrency):
    decimal_digits = 0


class JMD(BaseCurrency):
    pass


class JOD(BaseCurrency):
    decimal_digits = 3


class JPY(BaseCurrency):
    decimal_digits = 0


class KES(BaseCurrency):
    pass


class KGS(BaseCurrency):
    pass


class KHR(BaseCurrency):
    pass


class KMF(BaseCurrency):
    decimal_digits = 0


class KPW(BaseCurrency):
    pass


class KRW(BaseCurrency):
    decimal_digits = 0


class KWD(BaseCurrency):
    decimal_digits = 3


class KYD(BaseCurrency):
    pass


class KZT(BaseCurrency):
    pass


class LAK(BaseCurrency):
    pass


class LBP(BaseCurrency):
    pass


class LKR(BaseCurrency):
    pass


class LRD(BaseCurrency):
    pass


class LSL(BaseCurrency):
    pass


class LYD(BaseCurrency):
    decimal_digits = 3


class MAD(BaseCurrency):
    pass


class MDL(BaseCurrency):
    pass


class MGA(BaseCurrency):
    pass


class MKD(BaseCurrency):
    pass


class MMK(BaseCurrency):
    pass


class MNT(BaseCurrency):
    pass


class MOP(BaseCurrency):
    pass


class MRU(BaseCurrency):
    pass


class MUR(BaseCurrency):
    pass


class MVR(BaseCurrency):
    pass


class MWK(BaseCurrency):
    pass


class MXN(BaseCurrency):
    pass


class MXV(BaseCurrency):
    pass


class MYR(BaseCurrency):
    pass


class MZN(BaseCurrency):
    pass


class NAD(BaseCurrency):
    pass


class NGN(BaseCurrency):
    pass


class NIO(BaseCurrency):
    pass


class NOK(BaseCurrency):
    pass


class NPR(BaseCurrency):
    pass


class NZD(BaseCurrency):
    pass


class OMR(BaseCurrency):
    decimal_digits = 3


class PAB(BaseCurrency):
    pass


class PEN(BaseCurrency):
    pass


class PGK(BaseCurrency):
    pass


class PHP(BaseCurrency):
    pass


class PKR(BaseCurrency):
    pass


class PLN(BaseCurrency):
    pass


class PYG(BaseCurrency):
    decimal_digits = 0


class QAR(BaseCurrency):
    pass


class RON(BaseCurrency):
    pass


class RSD(BaseCurrency):
    pass


class RUB(BaseCurrency):
    pass


class RWF(BaseCurrency):
    decimal_digits = 0


class SAR(BaseCurrency):
    pass


class SBD(BaseCurrency):
    pass


class SCR(BaseCurrency):
    pass


class SDG(BaseCurrency):
    pass


class SEK(BaseCurrency):
    pass


class SGD(BaseCurrency):
    pass


class SHP(BaseCurrency):
    pass


class SLL(BaseCurrency):
    pass


class SOS(BaseCurrency):
    pass


class SRD(BaseCurrency):
    pass


class SSP(BaseCurrency):
    pass


class STN(BaseCurrency):
    pass


class SVC(BaseCurrency):
    pass


class SYP(BaseCurrency):
    pass


class SZL(BaseCurrency):
    pass


class THB(BaseCurrency):
    pass


class TJS(BaseCurrency):
    pass


class TMT(BaseCurrency):
    pass


class TND(BaseCurrency):
    decimal_digits = 3


class TOP(BaseCurrency):
    pass


class TRY(BaseCurrency):
    pass


class TTD(BaseCurrency):
    pass


class TWD(BaseCurrency):
    interchangeable_with = ("NTD",)


class TZS(BaseCurrency):
    pass


class UAH(BaseCurrency):
    pass


class UGX(BaseCurrency):
    decimal_digit = 0


class USD(BaseCurrency):
    pass


class USN(BaseCurrency):
    pass


class UYI(BaseCurrency):
    decimal_digits = 0


class UYU(BaseCurrency):
    pass


class UYW(BaseCurrency):
    decimal_digits = 4


class UZS(BaseCurrency):
    pass


class VES(BaseCurrency):
    pass


class VND(BaseCurrency):
    decimal_digits = 0


class VUV(BaseCurrency):
    decimal_digits = 0


class WST(BaseCurrency):
    pass


class XAF(BaseCurrency):
    decimal_digits = 0


class XAG(BaseCurrency):
    pass


class XAU(BaseCurrency):
    pass


class XBA(BaseCurrency):
    pass


class XBB(BaseCurrency):
    pass


class XBC(BaseCurrency):
    pass


class XBD(BaseCurrency):
    pass


class XCD(BaseCurrency):
    pass


class XDR(BaseCurrency):
    pass


class XOF(BaseCurrency):
    decimal_digits = 0


class XPD(BaseCurrency):
    pass


class XPF(BaseCurrency):
    decimal_digits = 0


class XPT(BaseCurrency):
    pass


class XSU(BaseCurrency):
    pass


class XTS(BaseCurrency):
    pass


class XUA(BaseCurrency):
    pass


class XXX(BaseCurrency):
    pass


class YER(BaseCurrency):
    pass


class ZAR(BaseCurrency):
    pass


class ZMW(BaseCurrency):
    pass


class ZWL(BaseCurrency):
    pass


# Unofficial currency codes


class CNH(BaseCurrency):
    interchangeable_with = ("CNY", "RMB")
    preferred_ticker = "CNY"


class GGP(BaseCurrency):
    pass


class IMP(BaseCurrency):
    pass


class JED(BaseCurrency):
    pass


class KID(BaseCurrency):
    pass


class NIS(BaseCurrency):
    interchangeable_with = ("ILS",)
    preferred_ticker = "ILS"


class NTD(BaseCurrency):
    interchangeable_with = ("TWD",)
    preferred_ticker = "TWD"


class PRB(BaseCurrency):
    pass


class SLS(BaseCurrency):
    pass


class RMB(BaseCurrency):
    interchangeable_with = ("CNH", "RMB")
    preferred_ticker = "CNY"


class TVD(BaseCurrency):
    pass


class ZWB(BaseCurrency):
    pass


# Historical currency codes


class ADF(BaseCurrency):
    pass


class ADP(BaseCurrency):
    decimal_digits = 0


class AFA(BaseCurrency):
    pass


class AOK(BaseCurrency):
    decimal_digits = 0


class AON(BaseCurrency):
    decimal_digits = 0


class AOR(BaseCurrency):
    decimal_digits = 0


class ARL(BaseCurrency):
    pass


class ARP(BaseCurrency):
    pass


class ARA(BaseCurrency):
    pass


class ATS(BaseCurrency):
    pass


class AZM(BaseCurrency):
    decimal_digits = 0


class BAD(BaseCurrency):
    pass


class BEF(BaseCurrency):
    pass


class BGL(BaseCurrency):
    pass


class BOP(BaseCurrency):
    pass


class BRB(BaseCurrency):
    pass


class BRC(BaseCurrency):
    pass


class BRN(BaseCurrency):
    pass


class BRE(BaseCurrency):
    pass


class BRR(BaseCurrency):
    pass


class BYB(BaseCurrency):
    pass


class BYR(BaseCurrency):
    decimal_digits = 0


class CSD(BaseCurrency):
    pass


class CSK(BaseCurrency):
    pass


class CYP(BaseCurrency):
    pass


class DDM(BaseCurrency):
    pass


class DEM(BaseCurrency):
    pass


class ECS(BaseCurrency):
    decimal_digits = 0


class ECV(BaseCurrency):
    pass


class EEK(BaseCurrency):
    pass


class ESA(BaseCurrency):
    pass


class ESB(BaseCurrency):
    pass


class ESP(BaseCurrency):
    decimal_digits = 0


class FIM(BaseCurrency):
    pass


class FRF(BaseCurrency):
    pass


class GNE(BaseCurrency):
    pass


class GHC(BaseCurrency):
    decimal_digits = 0


class GQE(BaseCurrency):
    pass


class GRD(BaseCurrency):
    pass


class GWP(BaseCurrency):
    pass


class HRD(BaseCurrency):
    pass


class IEP(BaseCurrency):
    pass


class ILP(BaseCurrency):
    decimal_digits = 3


class ILR(BaseCurrency):
    pass


class ISJ(BaseCurrency):
    pass


class ITL(BaseCurrency):
    decimal_digits = 0


class LAJ(BaseCurrency):
    pass


class LTL(BaseCurrency):
    pass


class LUF(BaseCurrency):
    pass


class LVL(BaseCurrency):
    pass


class MAF(BaseCurrency):
    pass


class MCF(BaseCurrency):
    pass


class MGF(BaseCurrency):
    pass


class MKN(BaseCurrency):
    pass


class MLF(BaseCurrency):
    pass


class MVQ(BaseCurrency):
    pass


class MRO(BaseCurrency):
    pass


class MXP(BaseCurrency):
    pass


class MZM(BaseCurrency):
    decimal_digits = 0


class MTL(BaseCurrency):
    pass


class NIC(BaseCurrency):
    pass


class NLG(BaseCurrency):
    pass


class PEH(BaseCurrency):
    pass


class PEI(BaseCurrency):
    pass


class PLZ(BaseCurrency):
    pass


class PTE(BaseCurrency):
    decimal_digits = 0


class ROL(BaseCurrency):
    pass


class RUR(BaseCurrency):
    pass


class SDD(BaseCurrency):
    decimal_digits = 0


class SDP(BaseCurrency):
    pass


class SIT(BaseCurrency):
    pass


class SKK(BaseCurrency):
    pass


class SML(BaseCurrency):
    decimal_digits = 0


class SRG(BaseCurrency):
    pass


class STD(BaseCurrency):
    pass


class SUR(BaseCurrency):
    pass


class TJR(BaseCurrency):
    pass


class TMM(BaseCurrency):
    decimal_digits = 0


class TPE(BaseCurrency):
    pass


class TRL(BaseCurrency):
    decimal_digits = 0


class UAK(BaseCurrency):
    pass


class UGS(BaseCurrency):
    pass


class USS(BaseCurrency):
    pass


class UYP(BaseCurrency):
    pass


class UYN(BaseCurrency):
    pass


class VAL(BaseCurrency):
    decimal_digits = 0


class VEB(BaseCurrency):
    pass


class VEF(BaseCurrency):
    pass


class XEU(BaseCurrency):
    pass


class XFO(BaseCurrency):
    pass


class XFU(BaseCurrency):
    pass


class YDD(BaseCurrency):
    pass


class YUD(BaseCurrency):
    pass


class YUN(BaseCurrency):
    pass


class YUR(BaseCurrency):
    pass


class YUO(BaseCurrency):
    pass


class YUG(BaseCurrency):
    pass


class YUM(BaseCurrency):
    pass


class ZAL(BaseCurrency):
    pass


class ZMK(BaseCurrency):
    pass


class ZRZ(BaseCurrency):
    decimal_digits = 3


class ZRN(BaseCurrency):
    pass


class ZWC(BaseCurrency):
    pass


class ZWD(BaseCurrency):
    pass


class ZWN(BaseCurrency):
    pass


class ZWR(BaseCurrency):
    pass


# Cryptocurrencies


class Bitcoin(BaseCurrency):
    ticker = "BTC"


BTC = Bitcoin
XBT = Bitcoin


class Ethereum(BaseCurrency):
    ticker = "ETH"


ETH = Ethereum


class XRP(BaseCurrency):
    ticker = "XRP"


class Tether(BaseCurrency):
    ticker = "USDT"


USDT = Tether


class USDCoin(BaseCurrency):
    ticker = "USDC"


CoinbaseUSDC = USDCoin
USDC = USDCoin


class BitcoinCash(BaseCurrency):
    ticker = "BCH"


BCH = BitcoinCash
XCH = BitcoinCash


class LiteCoin(BaseCurrency):
    ticker = "LTC"


LTC = LiteCoin


class EOS(BaseCurrency):
    ticker = "EOS"


class BinanceCoin(BaseCurrency):
    ticker = "BNB"


BNB = BinanceCoin


class StellarLumen(BaseCurrency):
    ticker = "XLM"


XLM = StellarLumen


class Monero(BaseCurrency):
    ticker = "XMR"


XMR = Monero


class DogeCoin(BaseCurrency):
    ticker = "DOGE"


DOGE = DogeCoin


def get_currency(ticker: str) -> BaseCurrency:
    return cast(BaseCurrency, getattr(sys.modules[__name__], ticker, BaseCurrency(ticker)))


def all_currencies() -> List[str]:
    return [ticker for ticker in dir(sys.modules[__name__]) if ticker and ticker == ticker.upper()]


class Currency(type):
    def __new__(cls, *args: Any, **kwargs: Any) -> BaseCurrency:  # type: ignore
        result: BaseCurrency = BaseCurrency(*args, **kwargs)
        return result

    @staticmethod
    def _load_currencies(currencies: List[str]) -> None:
        for ticker in currencies:
            setattr(Currency, ticker, getattr(sys.modules[__name__], ticker))


Currency._load_currencies(all_currencies())



class MoneyException(Exception):
    pass


class CurrencyMismatchError(MoneyException, ValueError):
    pass


class ConversionError(MoneyException, ValueError):
    pass


class InvalidOperandError(MoneyException, TypeError):
    pass



from .__version__ import __version__, __version_info__  # noqa
from .currency import BaseCurrency, Currency, get_currency  # noqa
from .money import Money  # noqa
from .exceptions import MoneyException, CurrencyMismatchError, ConversionError, InvalidOperandError  # noqa

__author__ = "Carl Oscar Aaro"
__email__ = "hello@carloscar.com"



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['stockholm']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'stockholm',
    'version': '0.3.2',
    'description': 'An up to date human friendly and flexible approach for development with any kind of monetary amounts',
    'long_description': '# `stockholm` — `Money` for Python 3\n[![pypi](https://badge.fury.io/py/stockholm.svg)](https://pypi.python.org/pypi/stockholm/)\n[![Made with Python](https://img.shields.io/pypi/pyversions/stockholm)](https://www.python.org/)\n[![Type hinted - mypy validated](https://img.shields.io/badge/typehinted-yes-teal)](https://github.com/kalaspuff/stockholm)\n[![MIT License](https://img.shields.io/github/license/kalaspuff/stockholm.svg)](https://github.com/kalaspuff/stockholm/blob/master/LICENSE)\n[![Code coverage](https://codecov.io/gh/kalaspuff/stockholm/branch/master/graph/badge.svg)](https://codecov.io/gh/kalaspuff/stockholm/tree/master/stockholm)\n\n*Library for formatting and performing arithmetic and comparison operations on monetary amounts. Also with support for currency handling, exchange and network transport structure generation as well as parsing.*\n\nAn up to date human friendly and flexible approach for development with any kind of monetary amounts.\n\nAt its bone a `Money` class for Python 3.x. This is a library to be used by backend and frontend API coders of fintech companies, web merchants and subscription services. A simple, yet powerful way of coding with money.\n\n---\n\n#### `from stockholm import Money`\nThe `stockholm.Money` object has full arithmetic support together with `int`, `float`, `Decimal`, other `Money` objects as well as `string`. The `stockholm.Money` object also supports complex string formatting functionality for easy debugging and a clean coding pattern.\n\n#### `from stockholm import Currency`\nCurrencies to monetary amounts can be specified using either currencies built with the `stockholm.Currency` metaclasses or simply by specifying the currency ticker as a string (for example `"SEK"` or `"EUR"`) when creating a new `Money` object.\n\nCurrencies using the `stockholm.Currency` metaclasses can hold additional options, such as default number of decimals in string output. Note that the amounts are usually never behind the scenes and uses the same precision and backend as `Decimal` values and can as well be interchangable with such values.\n\n\n## Installation with `pip`\nLike you would install any other Python package, use `pip`, `poetry`, `pipenv` or your weapon of choice.\n```\n$ pip install stockholm\n```\n\n\n## Examples\n\n#### Arithmetics - fully supported\n*Full arithmetic support with different types, backed by `Decimal` for dealing with rounding errors, while also keeping the monetary amount fully currency aware.*\n```python\nfrom stockholm import Money\n\nmoney = Money("4711.50", currency="SEK")\nprint(money)\n# 4711.50 SEK\n\noutput = (money + 100) * 3 + Money(50)\nprint(output)\n# 14484.50 SEK\n\nprint(output / 5)\n# 2896.90 SEK\n\nprint(round(output / 3, 4))\n# 4828.1667 SEK\n\nprint(round(output / 3, 1))\n# 4828.20 SEK\n```\n\n#### Formatting / Advanced string formatting\n*Advanced string formatting functionality.*\n```python\nfrom stockholm import Money\n\njpy_money = Money(1352953, "JPY")\nexchange_rate = Money("0.08861326")\nsek_money = Money(jpy_money * exchange_rate, "SEK")\n\nprint(f"I have {jpy_money:,.0m} which equals around {sek_money:,.2m}")\nprint(f"The exchange rate is {exchange_rate} ({jpy_money:c} -> {sek_money:c})")\n# I have 1,352,953 JPY which equals around 119,889.58 SEK\n# The exchange rate is 0.08861326 (JPY -> SEK)\n\n# Standard string format uses default min decimals up to 9 decimals\nprint(f"{sek_money}")  # 119889.57595678 SEK\n\n# Format type "f" works the same way as formatting a float or Decimal\nprint(f"{jpy_money:.0f}")  # 1352953\nprint(f"{sek_money:.2f}")  # 119889.58\nprint(f"{sek_money:.1f}")  # 119889.6\nprint(f"{sek_money:.0f}")  # 119890\n\n# Format type "m" works as "f" but includes the currency in string output\nprint(f"{sek_money:.2m}")  # 119889.57 SEK\nprint(f"{sek_money:.4m}")  # 119889.5760 SEK\nprint(f"{sek_money:+,.4m}")  # +119,889.5760 SEK\n\n# An uppercase "M" puts the currency ticker in front of the amount\nprint(f"{sek_money:.4M}")  # SEK 119889.5760\n\n# Format type "c" will just output the currency used in the monetary amount\nprint(f"{sek_money:c}")  # SEK\n```\n\n*Use `stockholm.Currency` types for proper defaults of minimum number of decimal digits to output in strings, etc. All ISO 4217 currency codes implemented, see https://github.com/kalaspuff/stockholm/blob/master/stockholm/currency.py for the full list.*\n```python\nfrom stockholm import Currency, Money, get_currency\nfrom stockholm.currency import JPY, SEK, EUR, IQD, USDCoin, Bitcoin\n\n# Most currencies has a minimum default digits set to 2 in strings\nprint(Money(4711, SEK))  # 4711.00 SEK\nprint(Money(4711, EUR))  # 4711.00 EUR\n\n# The stockholm.currency.JPY has a minimum default digits set to 0\nprint(Money(4711, JPY))  # 4711 JPY\n\n# Some currencies even has a minimum default of 3 or 4 digits\nprint(Money(4711, IQD))  # 4711.000 IQD\n\n# Some complex non ISO 4217 currencies, assets or tokens may define\n# their own ticker, for example a "USD Coin" uses the ticker "USDC"\nprint(Money(4711, USDCoin))  # 4711.00 USDC\nprint(Money(4711, Bitcoin))  # 4711.00 BTC\n\n# You can also use the shorthand stockholm.Currency object which\n# holds all ISO 4217 three character codes as objects.\nprint(Money(1338, Currency.JPY))  # 1338 JPY\n\n# or call the get_currency function\nprint(Money(1338, get_currency("JPY")))  # 1338 JPY\n\n```\n\n#### Input data types in flexible variants\n*Flexible ways for assigning values to a monetary amount using many different input data types and methods.*\n```python\nfrom decimal import Decimal\nfrom stockholm import Money\n\nMoney(100, currency="EUR")\n# <stockholm.Money: "100.00 EUR">\n\nMoney("1338 USD")\n# <stockholm.Money: "1338.00 USD">\n\nMoney("0.5")\n# <stockholm.Money: "0.50">\n\namount = Decimal(5000) / 3\nMoney(amount, currency="XDR")\n# <stockholm.Money: "1666.666666667">\n\nmoney = Money("0.30285471")\nMoney(money, currency="BTC")\n# <stockholm.Money: "0.30285471 BTC">\n\ncents_as_str = "471100"\nMoney(cents_as_str, currency="USD", from_sub_units=True)\n# <stockholm.Money: "4711.00 USD">\n```\n\n#### List arithmetics - summary of monetary amounts in list\n*Adding several monetary amounts from a list.*\n```python\nfrom stockholm import Money\n\namounts = [\n    Money(1),\n    Money("1.50"),\n    Money("1000"),\n]\n\n# Use Money.sum to deal with complex values of different data types\nMoney.sum(amounts)\n# <stockholm.Money: "1002.50">\n\n# Built-in sum may also be used (if only working with monetary amounts)\nsum(amounts)\n# <stockholm.Money: "1002.50">\n```\n\n#### Conversion for other transport medium (for example Protocol Buffers)\n*Easily splittable into `units` and `nanos` for transport in network medium, for example using the [`google.type.Money` protobuf definition](https://github.com/googleapis/googleapis/blob/master/google/type/money.proto) when using Protocol Buffers.*\n```python\nfrom stockholm import Money\n\nmoney = Money("22583.75", "SEK")\nmoney.units, money.nanos, str(money.currency)\n# (22583, 750000000, \'SEK\')\n\n# or vice versa\nMoney(units=22583, nanos=750000000, currency="SEK")\n# <stockholm.Money: "22583.75 SEK">\n```\n\n\n## Acknowledgements\nBuilt with inspiration from https://github.com/carlospalol/money and https://github.com/vimeo/py-money\n',
    'author': 'Carl Oscar Aaro',
    'author_email': 'hello@carloscar.com',
    'url': 'https://github.com/kalaspuff/stockholm',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
